@@ -1,9 +1,14 @@
+#!./runmodule.sh
+
 import enum
 import os
+import stat
 import filecmp
 import typing
 
-from .config import Config, DictConfig, MultiConfig, MultiDictConfig, ConfigId
+from confattr import Config, DictConfig, MultiConfig, MultiDictConfig, ConfigId
+
+from .symlink import SYMLINK_TYPE, are_symlinks_supported, is_symlink, read_symlink, abspath, is_internal_link, config_change_abs_internal_symlink_to_target
 
 
 # ---------- types ----------
@@ -14,11 +19,14 @@ from .config import Config, DictConfig, MultiConfig, MultiDictConfig, ConfigId
 class TYPE(enum.Enum):
 	FILE = 'f'
 	DIRECTORY = 'd'
+	LINK = 'l'
 	NOT_EXISTING = 'x'
 
 	@classmethod
 	def from_path(cls, path: str) -> 'TYPE':
-		if not os.path.exists(path):
+		if is_symlink(path):
+			return TYPE.LINK
+		elif not os.path.exists(path):
 			return TYPE.NOT_EXISTING
 		elif os.path.isdir(path):
 			return TYPE.DIRECTORY
@@ -35,6 +43,7 @@ class STATE(enum.Enum):
 	NEW      = '+'
 	DELETED  = '-'
 	SAME     = '='
+	UNKNOWN  = '?'
 
 	# ComparisonNode only
 	NEWER    = '>'
@@ -87,6 +96,7 @@ DIRECTION_OF_CHILDREN = typing.Union[DIRECTION, MIXED_DIRECTION]
 class ACTION(enum.Enum):
 	NONE        = 'none'
 	IGNORE      = 'ignore'
+	ERROR       = 'error'
 	CREATE      = 'create'
 	DELETE      = 'delete'
 	UNDO_CREATE = 'undo-create'
@@ -112,6 +122,17 @@ class ACTION(enum.Enum):
 
 
 
+# ---------- compare mode ----------
+
+@enum.unique
+class CMP(enum.Enum):
+	SHALLOW = 'shallow'
+	MIXED = 'mixed'
+	DEEP = 'deep'
+
+
+# ---------- node objects ----------
+
 class CommandNotAllowed(Exception):
 	pass
 
@@ -120,8 +141,21 @@ class ComparisonNode(object):
 
 	number_nodes = 0
 
-	expand_level = Config('diff.model.expand-level', 1,
+	expand_level = Config('diff.model.expand-level', 1, unit='',
 		help='0 means no directories are expanded by default, 1 means the topmost directories are expanded, 2 means the topmost directories and their direct subdirectories are expanded, ..., -1 means all directories are expanded')
+	compare_mode = Config('diff.model.compare-mode', CMP.SHALLOW, help = {
+		CMP.SHALLOW : 'Compare only file type, size and modification time. Do not compare the content. See also %diff.model.time-precision%.',
+		CMP.MIXED : 'Two files are taken to be identical if file type, size and modification time are identical. Compare the content if the modification time differs. This uses filecmp.cmp(shallow=True).',
+		CMP.DEEP : 'Compare the content except if the file size differs. This uses filecmp.cmp(shallow=False).',
+
+	})
+	time_precision = Config('diff.model.time-precision', 2.0,
+		unit = 'seconds',
+		help = 'Different file system types have different resolutions for time stamps. Therefore two different time stamps can refer to the same point in time. Two time stamps are considered unequal if their difference is bigger than this value. This value is considered only if %diff.model.compare-mode% is shallow.')
+	is_absolute_symlink_same_as_relative_symlink = Config('symlink.absolute-equals-relative', True, help='Whether two sym links can be considered to be the same if one of them is an absolute link and the other a relative link')
+
+	change_abs_internal_symlink_to_target = config_change_abs_internal_symlink_to_target
+
 
 	# ---------- properties ----------
 
@@ -151,9 +185,11 @@ class ComparisonNode(object):
 		path_src: str,
 		path_dst: str,
 		parent: typing.Optional['DirectoryComparisonNode'] = None,
+		root : typing.Optional['ComparisonNode'] = None,  # I cannot ensure that root is a DirectoryComparisonNode because if the user is passing two files as arguments there is no DirectoryComparisonNode
 		config_id: typing.Optional[ConfigId] = None,
 		direction: typing.Optional[DIRECTION] = None,
 		level: int = 0,
+		symlinks_supported: typing.Optional[bool] = None,
 	) -> None:
 		'''
 		path_src    path_dst
@@ -166,6 +202,8 @@ class ComparisonNode(object):
 		   ↘     ↓
 		       action     → output to user
 
+		path_src and path_dst are not allowed to end on a slash.
+
 		The state describes how the src has changed since the last backup (represented by dst).
 		The action describes how dst needs to be changed in order to make it the same like src.
 
@@ -190,6 +228,7 @@ class ComparisonNode(object):
 		if config_id is None:
 			config_id = MultiConfig.default_config_id
 		self.parent   = parent
+		self.root     = root if root is not None else self
 		self.name     = name
 		self.path_src = path_src
 		self.path_dst = path_dst
@@ -204,12 +243,12 @@ class ComparisonNode(object):
 		self.action: ACTION
 		self.listener_direction_changed: typing.Optional[typing.Callable[[], None]] = None
 		self.level = level
+		self.symlinks_supported = are_symlinks_supported((self.root.path_src, self.root.path_dst)) if symlinks_supported is None else symlinks_supported
 
 		if direction is not None:
 			self.direction = direction
 
-		type(self).number_nodes += 1
-		self.update()
+		self.update(update_parent=False)
 
 
 	# ---------- update ----------
@@ -230,7 +269,7 @@ class ComparisonNode(object):
 		STATE.DELETED : DIRECTION.SRC_TO_DST,
 		STATE.NEWER   : DIRECTION.SRC_TO_DST,
 		STATE.OLDER   : DIRECTION.SRC_TO_DST,
-	}, ignore_keys={STATE.MODIFIED_DIR}, allowed_values=(DIRECTION.SRC_TO_DST, DIRECTION.DST_TO_SRC, DIRECTION.NONE))
+	}, ignore_keys={STATE.MODIFIED_DIR, STATE.UNKNOWN}, allowed_values=(DIRECTION.SRC_TO_DST, DIRECTION.DST_TO_SRC, DIRECTION.NONE))
 
 	direction_state_action_map = {
 		DIRECTION.SRC_TO_DST : {
@@ -259,9 +298,18 @@ class ComparisonNode(object):
 
 	# ---------- update ----------
 
-	def update(self) -> None:
-		self.type_src = TYPE.from_path(self.path_src)
-		self.type_dst = TYPE.from_path(self.path_dst)
+	def update(self, *, update_parent: bool = True) -> None:
+		self.error_src = None
+		self.error_dst = None
+
+		if self.is_child_of_link_src():
+			self.type_src = TYPE.NOT_EXISTING
+		else:
+			self.type_src = TYPE.from_path(self.path_src)
+		if self.is_child_of_link_dst():
+			self.type_dst = TYPE.NOT_EXISTING
+		else:
+			self.type_dst = TYPE.from_path(self.path_dst)
 
 		if self.type_src is TYPE.DIRECTORY or self.type_dst is TYPE.DIRECTORY:
 			self.convert_to_directory_node()
@@ -278,6 +326,26 @@ class ComparisonNode(object):
 			self.type_specific_update()
 
 		self.update_direction()
+		ComparisonNode.number_nodes += 1
+
+		if update_parent:
+			self.update_parent_recursively()
+
+	def is_child_of_link_src(self) -> bool:
+		if self.parent is None:
+			return False
+		elif self.parent.type_src is TYPE.LINK:
+			return True
+		else:
+			return self.parent.is_child_of_link_src()
+
+	def is_child_of_link_dst(self) -> bool:
+		if self.parent is None:
+			return False
+		elif self.parent.type_dst is TYPE.LINK:
+			return True
+		else:
+			return self.parent.is_child_of_link_dst()
 
 	def convert_to_directory_node(self) -> None:
 		self.__class__ = DirectoryComparisonNode
@@ -291,7 +359,14 @@ class ComparisonNode(object):
 		pass
 
 	def type_specific_update(self) -> None:
-		if filecmp.cmp(self.path_src, self.path_dst):
+		if self.type_src is TYPE.LINK or self.type_dst is TYPE.LINK:
+			if self.links_equal(self.path_src, self.path_dst, self.root.path_src, self.root.path_dst):
+				state = STATE.SAME
+			elif os.stat(self.path_src, follow_symlinks=False).st_mtime < os.stat(self.path_dst, follow_symlinks=False).st_mtime:
+				state = STATE.OLDER
+			else:
+				state = STATE.NEWER
+		elif self.files_equal(self.path_src, self.path_dst):
 			state = STATE.SAME
 		elif os.path.getmtime(self.path_src) < os.path.getmtime(self.path_dst):
 			state = STATE.OLDER
@@ -299,6 +374,57 @@ class ComparisonNode(object):
 			state = STATE.NEWER
 		self.state = state
 
+	def links_equal(self, src: str, dst: str, root_src: str, root_dst: str) -> bool:
+		try:
+			t1, p1 = read_symlink(src)
+			t2, p2 = read_symlink(dst)
+		except (OSError, FileNotFoundError):
+			return False
+		if not self.is_absolute_symlink_same_as_relative_symlink and t1 is not t2:
+			return False
+		# I cannot use os.path.samefile here because (a) it's not guaranteed that the files exist (b) they are not supposed to link to the same file. p1 points to src, p2 to dst
+
+		# convert links to relative links so that they compare to be equal
+		# even if one of them is absolute and the other is relative
+		# and even if dst is an absolute link which has been changed to point to the dst side
+		if t1 is SYMLINK_TYPE.ABSOLUTE:
+			p1 = os.path.relpath(p1, os.path.dirname(src))
+		if t2 is SYMLINK_TYPE.ABSOLUTE:
+			if self.change_abs_internal_symlink_to_target and is_internal_link(link=src, target=p1, root=root_src):
+				ref = dst
+			else:
+				ref = src
+			p2 = os.path.relpath(p2, os.path.dirname(ref))
+
+		return p1 == p2
+
+	def files_equal(self, src: str, dst: str) -> bool:
+		if self.compare_mode is CMP.SHALLOW:
+			return self.files_equal_shallow(src, dst)
+		elif self.compare_mode is CMP.MIXED:
+			return filecmp.cmp(self.path_src, self.path_dst)
+		elif self.compare_mode is CMP.DEEP:
+			return filecmp.cmp(self.path_src, self.path_dst, shallow=False)
+		else:
+			assert False
+
+	def files_equal_shallow(self, src: str, dst: str) -> bool:
+		s1 = os.stat(src)
+		s2 = os.stat(dst)
+
+		if s1.st_size != s2.st_size:
+			return False
+
+		# check file type, I am ignoring the permissions part because not all file systems support all permissions
+		if stat.S_IFMT(s1.st_mode) != stat.S_IFMT(s2.st_mode):
+			return False
+
+		# avoid math if the modification time is equal for efficiency
+		if s1.st_mtime == s2.st_mtime:
+			return True
+
+		return abs(s1.st_mtime - s2.st_mtime) <= self.time_precision
+
 	def update_direction(self) -> None:
 		self.default_direction = self.state_direction_map[self.state]
 		if not hasattr(self, 'direction'):
@@ -306,6 +432,39 @@ class ComparisonNode(object):
 		self.update_action()
 
 	def update_action(self) -> None:
+		if self.direction is not DIRECTION.NONE and (self.type_src is TYPE.LINK or self.type_dst is TYPE.LINK):
+			is_broken_src = self.type_src is TYPE.LINK and not os.path.exists(self.path_src)
+			is_broken_dst = self.type_dst is TYPE.LINK and not os.path.exists(self.path_dst)
+			BROKEN_LINK = 'broken link'
+			if is_broken_src and is_broken_dst:
+				self.error_src = BROKEN_LINK
+				self.error_dst = BROKEN_LINK
+				self.action = ACTION.ERROR
+				return
+			elif self.state is STATE.SAME:
+				# if the links are equal and one of them is existing
+				# they are both ok even if the target does not exist on the backup side
+				pass
+			elif is_broken_src:
+				self.error_src = BROKEN_LINK
+				if self.direction is DIRECTION.SRC_TO_DST:
+					self.action = ACTION.ERROR
+					return
+			elif is_broken_dst:
+				self.error_dst = BROKEN_LINK
+				if self.direction is DIRECTION.DST_TO_SRC:
+					self.action = ACTION.ERROR
+					return
+		if not self.symlinks_supported:
+			if self.type_src is TYPE.LINK and self.direction is DIRECTION.SRC_TO_DST:
+				self.error_dst = 'file system does not support sym links'
+				self.action = ACTION.ERROR
+				return
+			elif self.type_dst is TYPE.LINK and self.direction is DIRECTION.DST_TO_SRC:
+				self.error_src = 'file system does not support sym links'
+				self.action = ACTION.ERROR
+				return
+
 		self.action = self.direction_state_action_map[self.direction][self.state]
 
 
@@ -408,15 +567,28 @@ class ComparisonNode(object):
 			return False
 		return self.direction is not self.default_direction
 
+	def has_error(self) -> bool:
+		return bool(self.error_src) or bool(self.error_dst)
+
+
+	# ---------- sort ----------
+
+	@classmethod
+	def sort_key_name(cls, name: str) -> str:
+		# In reality I mean Union[SupportsDunderLT, SupportsDunderGT] as return type, as mypy suggests.
+		# It seems, however, that neither SupportsDunderLT nor SupportsDunderGT are defined.
+		# And for defining them myself I would need Protocol, which is introduced in Python 3.8.
+		# But I don't want to increase the required Python version just for this type hint.
+		# So feel free to change the return type as needed.
+		# E.g. for correct sorting of numbers it would make more sense to return a tuple.
+		return name
+
 
 	# ---------- to string ----------
 
 	def __repr__(self) -> str:
 		return '<%s object %s>' % (type(self).__name__, self.path_src)
 
-	def __str__(self) -> str:
-		return self.tostring(0)
-
 	def tostring(self, level: int) -> str:
 		width_left = 30
 		indentation = level * '    '
@@ -457,6 +629,8 @@ class ComparisonNode(object):
 
 class DirectoryComparisonNode(ComparisonNode):
 
+	auto_load_children = Config('diff.model.auto-load-children', True, help='turning this off saves time while scanning for changes when large directories have been created, deleted or renamed. This comes at the price that the contents of these directories are not included in the statistics. That means if you remove a directory and create a file with the same name the statistics will only show that a type will be changed, they will not show that the contents of the directory will be deleted. Expanding a not preloaded directory is possible but the time which has been saved while scanning is then needed when expanding the directory.')
+
 	# ---------- mappings ----------
 
 	state_direction_map = ComparisonNode.state_direction_map
@@ -466,6 +640,12 @@ class DirectoryComparisonNode(ComparisonNode):
 	# this must be NONE because the directory exists on both sides, only the children differ
 	# if set to something else the resulting ACTION may include a direction in which the children are not changed
 	state_direction_map[STATE.MODIFIED_DIR] = DIRECTION.NONE
+	# this entry does not really make sense it's just so that I can use super().update_direction()
+	# I am distinguishing between direction=NONE => action=IGNORE and direction!=NONE => action=ERROR
+	# It does not make a difference whether a node with state=UNKNOWN has direction=SRC_TO_DST or DST_TO_SRC
+	# because the direction_of_children is always considered to be MIXED
+	# and the format for changed and unchanged direction are the same
+	state_direction_map[STATE.UNKNOWN] = DIRECTION.SRC_TO_DST
 
 	direction_state_action_map = ComparisonNode.direction_state_action_map
 	direction_state_action_map[DIRECTION.SRC_TO_DST][STATE.REPLACED_FILE_BY_DIRECTORY] = ACTION.CHANGE_DESTINATION_TYPE
@@ -477,6 +657,9 @@ class DirectoryComparisonNode(ComparisonNode):
 	direction_state_action_map[DIRECTION.NONE][STATE.REPLACED_FILE_BY_DIRECTORY]       = ACTION.IGNORE
 	direction_state_action_map[DIRECTION.NONE][STATE.REPLACED_DIRECTORY_BY_FILE]       = ACTION.IGNORE
 	direction_state_action_map[DIRECTION.NONE][STATE.MODIFIED_DIR]                     = ACTION.IGNORE
+	direction_state_action_map[DIRECTION.SRC_TO_DST][STATE.UNKNOWN]                    = ACTION.ERROR
+	direction_state_action_map[DIRECTION.DST_TO_SRC][STATE.UNKNOWN]                    = ACTION.ERROR
+	direction_state_action_map[DIRECTION.NONE][STATE.UNKNOWN]                          = ACTION.IGNORE
 
 
 	# ---------- update ----------
@@ -514,7 +697,7 @@ class DirectoryComparisonNode(ComparisonNode):
 		'''
 		self.loaded_children = True
 
-		if os.path.isdir(self.path_src):
+		if self.type_src is TYPE.DIRECTORY and not self.is_child_of_link_src() and os.path.isdir(self.path_src) and not self.is_symlink_circle(self.path_src):
 			try:
 				children_src = os.listdir(self.path_src)
 			except PermissionError as e:
@@ -523,7 +706,7 @@ class DirectoryComparisonNode(ComparisonNode):
 		else:
 			children_src = []
 
-		if os.path.isdir(self.path_dst):
+		if self.type_dst is TYPE.DIRECTORY and not self.is_child_of_link_dst() and os.path.isdir(self.path_dst) and not self.is_symlink_circle(self.path_dst):
 			try:
 				children_dst = os.listdir(self.path_dst)
 			except PermissionError as e:
@@ -536,15 +719,15 @@ class DirectoryComparisonNode(ComparisonNode):
 
 		i = 0
 		n = len(self.children)
-		iterator_filenames = iter(sorted(content))
+		iterator_filenames = iter(sorted(content, key=self.sort_key_name))
 		try:
 			filename = next(iterator_filenames)
 			while i < n:
 				child = self.children[i]
 				# Python 3 has no cmp anymore
 				if filename == child.name:
-					child.update()
-				elif filename < child.name:
+					child.update(update_parent=False)
+				elif self.sort_key_name(filename) < self.sort_key_name(child.name):
 					child = self.create_child_node(filename)
 					self.children.insert(i, child)
 					n += 1
@@ -553,7 +736,7 @@ class DirectoryComparisonNode(ComparisonNode):
 					n -= 1
 					continue
 
-				if self.state is STATE.SAME and child.state is not STATE.SAME:
+				if self.state is STATE.SAME and child.action is not ACTION.NONE:
 					self.state = STATE.MODIFIED_DIR
 				i += 1
 				filename = next(iterator_filenames)
@@ -564,12 +747,31 @@ class DirectoryComparisonNode(ComparisonNode):
 		except StopIteration:
 			del self.children[i:]
 
+		if self.has_error():
+			self.state = STATE.UNKNOWN
+
+	@classmethod
+	def is_symlink_circle(cls, path: str) -> bool:
+		if not is_symlink(path):
+			return False
+
+		path_list = path.split(os.path.sep)
+		for i in range(len(path_list) - 1, 0, -1):
+			p = os.path.sep.join(path_list[:i])
+			if not p:
+				p = os.path.sep
+			if os.path.samefile(p, path):
+				return True
+		return False
+
+
 	def create_child_node(self, filename: str) -> ComparisonNode:
 		direction = None
 		if self.state in STATES_WHICH_HAND_THEIR_DIRECTION_DOWN_TO_THEIR_CHILDREN:
 			direction = self.direction
 		return ComparisonNode(
 			parent = self,
+			root = self.root,
 			name = filename,
 			# this addition of strings is around 8x faster than os.path.join
 			path_src = self.path_src + os.sep + filename,
@@ -577,15 +779,20 @@ class DirectoryComparisonNode(ComparisonNode):
 			config_id = self.config_id,
 			direction = direction,
 			level = self.level + 1,
+			symlinks_supported = self.symlinks_supported,
 		)
 
 	def create_child_node_and_update_state(self, filename: str) -> ComparisonNode:
 		node = self.create_child_node(filename)
-		if self.state is STATE.SAME and node.state is not STATE.SAME:
+		if self.state is STATE.SAME and node.action is not ACTION.NONE:
 			self.state = STATE.MODIFIED_DIR
 		return node
 
 	def update_direction(self) -> None:
+		if not self.loaded_children and self.auto_load_children:
+			self.direction = self.state_direction_map[self.state]
+			self.update_children()
+
 		self.statistics.clear()
 		for child in self.children:
 			self.statistics.add(child)
@@ -706,6 +913,7 @@ class DirectoryComparisonNode(ComparisonNode):
 		if not self.is_expanded and not self.children:
 			self.update_children()
 			self.update_direction() # to update statistics
+			self.update_parent_recursively() # update parent statistics
 		self.is_expanded = value
 
 
@@ -732,17 +940,27 @@ class MetaNode(DirectoryComparisonNode):
 		self.path_dst = 'Destination'
 		self.type_src = TYPE.DIRECTORY
 		self.type_dst = TYPE.DIRECTORY
+		self.error_src = None
+		self.error_dst = None
 		self.state: STATE
 		self.default_direction = DIRECTION.NONE
 		self.direction = DIRECTION.NONE
 		self.action: ACTION
 		self.listener_direction_changed: typing.Optional[typing.Callable[[], None]] = None
 		self.level = -1
+		# symlinks_supported is checked in update_action to set an error if one side is a LINK and the file system on the other side does not support it
+		# but for a MetaNode both sides are a DIRECTORY by definition so it does not matter which value symlinks_supported has
+		self.symlinks_supported = True
 
 	def load(self, paths: typing.Sequence[typing.Tuple[ConfigId, str, str]]) -> None:
 		self.children = []
 		for config_id, src, dst in paths:
+			src = src.rstrip(os.path.sep)
+			dst = dst.rstrip(os.path.sep)
 			name = os.path.split(src)[0]
+			src = os.path.realpath(src)
+			dst = os.path.realpath(dst)
+			# yes, I don't want to set root to self. root is supposed to be the top most DirectoryComparisonNode, not the MetaNode.
 			self.children.append(ComparisonNode(name, src, dst, parent=self, config_id=config_id))
 
 		self.loaded_children = True
@@ -761,33 +979,32 @@ class MetaNode(DirectoryComparisonNode):
 			if child.state is not STATE.SAME:
 				self.state = STATE.MODIFIED_DIR
 
-	def update(self) -> None:
+	def update(self, *, update_parent: bool = True) -> None:
 		for child in self.children:
-			child.update()
+			child.update(update_parent=False)
 		self.update_state()
 		self.update_statistics()
+		if update_parent:
+			self.update_parent_recursively()
 
 
 class Statistics(object):
 
-	__slots__ = ('statistics', 'has_child_direction_been_changed', 'number_nodes_with_errors')
+	__slots__ = ('statistics', 'has_child_direction_been_changed')
 
 
 	def __init__(self) -> None:
 		self.statistics: typing.Dict[ACTION, int] = dict()
 		self.has_child_direction_been_changed = False
-		self.number_nodes_with_errors = 0
 
 	def clear(self) -> None:
 		self.statistics = {}
 		self.has_child_direction_been_changed = False
-		self.number_nodes_with_errors = 0
 
 	def add(self, other: ComparisonNode) -> None:
 		if isinstance(other, DirectoryComparisonNode):
 			for action, number in other.statistics.statistics.items():
 				self.statistics[action] = self.statistics.get(action, 0) + number
-			self.number_nodes_with_errors += other.statistics.number_nodes_with_errors
 
 			if other.has_child_direction_been_changed():
 				self.has_child_direction_been_changed = True
@@ -796,9 +1013,6 @@ class Statistics(object):
 		number = 1
 		self.statistics[action] = self.statistics.get(action, 0) + number
 
-		if other.error_src or other.error_dst:
-			self.number_nodes_with_errors += 1
-
 		if other.has_direction_been_changed():
 			self.has_child_direction_been_changed = True
 
@@ -815,6 +1029,7 @@ class Statistics(object):
 		ACTION.UNDO_DELETE_DIRECTORY_BUT_DELETE_SOME_CHILDREN,
 		ACTION.CHANGE_DESTINATION_TYPE_BUT_DELETE_SOME_CHILDREN,
 		ACTION.CHANGE_SOURCE_TYPE_BUT_DELETE_SOME_CHILDREN,
+		ACTION.ERROR,
 	)
 	ACTIONS_CHANGE_SOURCE = (
 		ACTION.UNDO_CREATE,
@@ -828,8 +1043,17 @@ class Statistics(object):
 		ACTION.UNDO_DELETE_DIRECTORY_BUT_DELETE_SOME_CHILDREN,
 		ACTION.CHANGE_DESTINATION_TYPE_BUT_DELETE_SOME_CHILDREN,
 		ACTION.CHANGE_SOURCE_TYPE_BUT_DELETE_SOME_CHILDREN,
+		ACTION.ERROR,
 	)
 
+	ACTIONS_NO_CHANGE = {
+		ACTION.NONE,
+		ACTION.IGNORE,
+		ACTION.DIR_CHANGE_DESTINATION,
+		ACTION.DIR_CHANGE_SOURCE,
+		ACTION.DIR_CHANGE_BOTH,
+	}
+
 	def get_direction(self) -> DIRECTION_OF_CHILDREN:
 		changed_source = any(True for a in self.ACTIONS_CHANGE_SOURCE if self.statistics.get(a, 0) > 0)
 		changed_destination = any(True for a in self.ACTIONS_CHANGE_DESTINATION if self.statistics.get(a, 0) > 0)
@@ -842,6 +1066,18 @@ class Statistics(object):
 		else:
 			return DIRECTION.NONE
 
+	def get_number_nodes_to_be_changed(self) -> int:
+		out = 0
+		for a in ACTION:
+			if a in self.ACTIONS_NO_CHANGE:
+				continue
+			out += self[a]
+
+		return out
+
+	def get_number_nodes_to_be_updated(self) -> int:
+		return sum(self.statistics.values())
+
 	def __getitem__(self, key: ACTION) -> int:
 		return self.statistics.get(key, 0)
 
@@ -868,4 +1104,4 @@ if __name__ == '__main__':
 		path_src = src,
 		path_dst = dst)
 
-	print('Tree: \n%s' % tree)
+	print('Tree: \n%s' % tree.tostring(0))
