@@ -1,10 +1,15 @@
-#!/usr/bin/env pytest
+#!../venv/bin/pytest -s
 
 import pytest
 import typing
 
-from udsync.config import Config, MultiConfig
+from confattr import Config, MultiConfig, ConfigFile
+import confattr.state
+
+from defs import fn_config, ui_callback
+
 from udsync.urwid_colors import Color, ColorConfig, ColorStr
+from udsync.urwid_multi_key_support import UrwidConfigFileArgparseCommand
 
 
 class COLOR:
@@ -68,6 +73,13 @@ def reset_config() -> None:
 	Config.instances.clear()
 	MultiConfig.config_ids.clear()
 	ColorConfig.color_configs.clear()
+	confattr.state.has_config_file_been_instantiated = False
+	confattr.state.has_any_config_file_been_instantiated = False
+
+def create_config_file() -> ConfigFile:
+	config_exporter = ConfigFile(appname='test', ignore_commands=[UrwidConfigFileArgparseCommand])
+	config_exporter.set_ui_callback(ui_callback)
+	return config_exporter
 
 
 # ------- Color -------
@@ -261,35 +273,42 @@ def test_colored_str_to_markup_str(registered_colors: typing.Set[str]) -> None:
 
 def test_colored_str_to_markup_one_color(registered_colors: typing.Set[str]) -> None:
 	assert ColorStr.to_markup('hello <color=red>world</color>') == ['hello ', ('red/default', 'world')]
-	assert {'red/default'} == registered_colors
+	assert {'red/default',
+		'red,standout/default'} == registered_colors
 
 def test_colored_str_to_markup_nested_colors(registered_colors: typing.Set[str]) -> None:
 	assert ColorStr.to_markup('<color=green>hello <color=red>world</color></color>') == ('green/default', ['hello ', ('red/default', 'world')])
-	assert {'green/default', 'red/default'} == registered_colors
+	assert {'green/default', 'red/default',
+		'green,standout/default', 'red,standout/default'} == registered_colors
 
 def test_colored_str_to_markup_chained_colors(registered_colors: typing.Set[str]) -> None:
 	assert ColorStr.to_markup('a<color=yellow>b<color=red>c</color>d<color=green>e</color>f</color>g') == ['a', ('yellow/default', ['b', ('red/default', 'c'), 'd', ('green/default', 'e'), 'f']), 'g']
-	assert {'yellow/default', 'red/default', 'green/default'} == registered_colors
+	assert {'yellow/default', 'red/default', 'green/default',
+		'yellow,standout/default', 'red,standout/default', 'green,standout/default'} == registered_colors
 
 def test_colored_str_to_markup_missing_closing_tag(registered_colors: typing.Set[str]) -> None:
 	assert ColorStr.to_markup('hello <color=red>world') == ['hello ', ('red/default', 'world')]
-	assert {'red/default'} == registered_colors
+	assert {'red/default',
+		'red,standout/default'} == registered_colors
 
 def test_colored_str_to_markup_missing_two_closing_tags(registered_colors: typing.Set[str]) -> None:
 	assert ColorStr.to_markup('<color=red>hello <color=green>world') == ('red/default', ['hello ', ('green/default', 'world')])
-	assert {'red/default', 'green/default'} == registered_colors
+	assert {'red/default', 'green/default',
+		'red,standout/default', 'green,standout/default'} == registered_colors
 
 def test_colored_str_to_markup_reference_other_color(registered_colors: typing.Set[str]) -> None:
 	ColorConfig('title', 'black/yellow')
 	assert ColorStr.to_markup('<color=red>hello <color={title}>world') == ('red/default', ['hello ', ('black/yellow', 'world')])
-	assert {'black/yellow', 'red/default'} == registered_colors
+	assert {'black/yellow', 'red/default',
+		'black,standout/yellow', 'red,standout/default',} == registered_colors
 
 def test_colored_str_to_markup_with_format(registered_colors: typing.Set[str]) -> None:
 	ColorConfig('O', 'black')
 	ColorConfig('I', 'red')
 	ColorConfig('II', 'green')
 	assert ColorStr.to_markup('{O}<color={I}>{I}<color={II}>{II}</color>{I}', format={'O':'blk', 'I':'red', 'II':'grn'}) == ['blk', ('red/default', ['red', ('green/default', 'grn'), 'red'])]
-	assert {'red/default', 'green/default'} == registered_colors
+	assert {'red/default', 'green/default',
+		'red,standout/default', 'green,standout/default'} == registered_colors
 
 def test_colored_str_to_markup_with_format_where_values_are_color_str(registered_colors: typing.Set[str]) -> None:
 	ColorConfig('O', 'black')
@@ -297,7 +316,8 @@ def test_colored_str_to_markup_with_format_where_values_are_color_str(registered
 	ColorConfig('II', 'green')
 	assert ColorStr.to_markup('{O}<color={I}>{I}<color={II}>{II}</color>{I}', format={'O':'blk', 'I':'r<color=default/red>e</color>d', 'II':'g<color=default/green>r</color>n'}, values_are_color_str=True) \
 		== ['blk', ('red/default', [['r', ('default/red', 'e'), 'd'], ('green/default', ['g', ('default/green', 'r'), 'n']), ['r', ('default/red', 'e'), 'd']])]
-	assert {'red/default', 'green/default', 'default/red', 'default/green'} == registered_colors
+	assert {'red/default', 'green/default', 'default/red', 'default/green',
+		'red,standout/default', 'green,standout/default', 'default,standout/red', 'default,standout/green'} == registered_colors
 
 
 # ------- simplify_markup -------
@@ -347,3 +367,139 @@ def test_simplify_markup__bug1() -> None:
 		(color_cmd, 'focus previous'),
 	]
 	assert ColorStr.simplify_markup(markup_in) == markup_expected
+
+
+# ------- save config -------
+
+def test_save_ignore_default_focus(fn_config: str) -> None:
+	class MyTestClass:
+		a = ColorConfig('a', 'red')
+
+	config_exporter = create_config_file()
+	config_exporter.save_file(fn_config, comments=False)
+
+	with open(fn_config, 'rt') as f:
+		assert f.read() == '''\
+set a = red/default
+'''
+
+def test_save_dont_ignore_default_focus(fn_config: str) -> None:
+	class MyTestClass:
+		a = ColorConfig('a', 'red', focus='yellow')
+
+	config_exporter = create_config_file()
+	config_exporter.save_file(fn_config, comments=False)
+
+	with open(fn_config, 'rt') as f:
+		assert f.read() == '''\
+set a = red/default
+set a-focus = yellow/default
+'''
+
+
+# ------- load config -------
+
+def test_load_set_focus_color(fn_config: str) -> None:
+	a = ColorConfig('a', 'default')
+	assert (a.color.attr_name, a.color.str_repr) == ('a', 'default/default')
+	assert (a.focus.color.attr_name, a.focus.color.str_repr) == ('a-focus', 'default,standout/default')
+
+	with open(fn_config, 'wt') as f:
+		f.write('''
+			set a-focus = red
+		''')
+
+	config_exporter = create_config_file()
+	config_exporter.load_file(fn_config)
+	assert (a.color.attr_name, a.color.str_repr) == ('a', 'default/default')
+	assert (a.focus.color.attr_name, a.focus.color.str_repr) == ('a-focus', 'red/default')
+
+def test_load_update_focus_color(fn_config: str) -> None:
+	a = ColorConfig('a', 'default')
+	assert (a.color.attr_name, a.color.str_repr) == ('a', 'default/default')
+	assert (a.focus.color.attr_name, a.focus.color.str_repr) == ('a-focus', 'default,standout/default')
+
+	with open(fn_config, 'wt') as f:
+		f.write('''
+			set a = cyan
+		''')
+
+	config_exporter = create_config_file()
+	config_exporter.load_file(fn_config)
+	assert (a.color.attr_name, a.color.str_repr) == ('a', 'cyan/default')
+	assert (a.focus.color.attr_name, a.focus.color.str_repr) == ('a-focus', 'cyan,standout/default')
+
+
+def test_load_dont_update_focus_color_if_it_was_given_in_init(fn_config: str) -> None:
+	a = ColorConfig('a', 'default', focus='black/yellow')
+	assert (a.color.attr_name, a.color.str_repr) == ('a', 'default/default')
+	assert (a.focus.color.attr_name, a.focus.color.str_repr) == ('a-focus', 'black/yellow')
+
+	with open(fn_config, 'wt') as f:
+		f.write('''
+			set a = cyan
+		''')
+
+	config_exporter = create_config_file()
+	config_exporter.load_file(fn_config)
+	assert (a.color.attr_name, a.color.str_repr) == ('a', 'cyan/default')
+	assert (a.focus.color.attr_name, a.focus.color.str_repr) == ('a-focus', 'black/yellow')
+
+def test_load_dont_update_focus_color_if_it_was_set_explicitly_in_config_file(fn_config: str) -> None:
+	a = ColorConfig('a', 'default')
+	assert (a.color.attr_name, a.color.str_repr) == ('a', 'default/default')
+	assert (a.focus.color.attr_name, a.focus.color.str_repr) == ('a-focus', 'default,standout/default')
+
+	with open(fn_config, 'wt') as f:
+		f.write('''
+			set a-focus = magenta/cyan
+			set a = cyan
+		''')
+
+	config_exporter = create_config_file()
+	config_exporter.load_file(fn_config)
+	assert (a.color.attr_name, a.color.str_repr) == ('a', 'cyan/default')
+	assert (a.focus.color.attr_name, a.focus.color.str_repr) == ('a-focus', 'magenta/cyan')
+
+
+# ------- save and load config -------
+
+def test_save_and_load_color(fn_config: str) -> None:
+	class MyTestClass:
+		color = ColorConfig('mycolor', 'default')
+
+	t = MyTestClass()
+	assert t.color == 'mycolor'
+	assert type(t).color.color.str_repr == 'default/default'
+
+	type(t).color.set_value(None, 'red/yellow')
+	config_exporter = create_config_file()
+	config_exporter.save_file(fn_config)
+	assert t.color == 'mycolor'
+	assert type(t).color.color.str_repr == 'red/yellow'
+
+	type(t).color.set_value(None, 'yellow/blue')
+	assert t.color == 'mycolor'
+	assert type(t).color.color.str_repr == 'yellow/blue'
+
+	config_exporter.load_file(fn_config)
+	assert t.color == 'mycolor'
+	assert type(t).color.color.str_repr == 'red/yellow'
+
+def test_save_and_load_color_str(fn_config: str) -> None:
+	class MyTestClass:
+		fmt = Config('fmt', ColorStr('hello world'))
+
+	t = MyTestClass()
+	assert t.fmt == 'hello world'
+
+	t.fmt = ColorStr('<color=green>hello</color> <color=blue>world</color>')
+	config_exporter = create_config_file()
+	config_exporter.save_file(fn_config)
+	assert t.fmt == '<color=green>hello</color> <color=blue>world</color>'
+
+	t.fmt = ColorStr('123')
+	assert t.fmt == '123'
+
+	config_exporter.load_file(fn_config)
+	assert t.fmt == '<color=green>hello</color> <color=blue>world</color>'
